@@ -1,13 +1,13 @@
 // -*- mode: js2; indent-tabs-mode: nil; js2-basic-offset: 4 -*-
-const Workspace = imports.ui.workspace;
+/* exported enable disable */
+const { Clutter } = imports.gi;
 
 const ExtensionUtils = imports.misc.extensionUtils;
-const Me = ExtensionUtils.getCurrentExtension();
-const Convenience = Me.imports.convenience;
+const Main = imports.ui.main;
+const { WindowPreview } = imports.ui.windowPreview;
+const Workspace = imports.ui.workspace;
 
 // testing settings for natural window placement strategy:
-const WINDOW_PLACEMENT_NATURAL_FILLGAPS = true;                     // enlarge windows at the end to fill gaps         // not implemented yet
-const WINDOW_PLACEMENT_NATURAL_GRID_FALLBACK = true;                // fallback to grid mode if all windows have the same size and positions.     // not implemented yet
 const WINDOW_PLACEMENT_NATURAL_ACCURACY = 20;                       // accuracy of window translate moves  (KDE-default: 20)
 const WINDOW_PLACEMENT_NATURAL_GAPS = 5;                            // half of the minimum gap between windows
 const WINDOW_PLACEMENT_NATURAL_MAX_TRANSLATIONS = 5000;             // safety limit for preventing endless loop if something is wrong in the algorithm
@@ -17,29 +17,25 @@ class Rect {
         [this.x, this.y, this.width, this.height] = [x, y, width, height];
     }
 
-    /**
-     * used in _calculateWindowTransformationsNatural to replace Meta.Rectangle that is too slow.
-     */
+    // used in _calculateWindowTransformationsNatural to replace Meta.Rectangle that is too slow.
     copy() {
         return new Rect(this.x, this.y, this.width, this.height);
     }
 
     union(rect2) {
         let dest = this.copy();
-        if (rect2.x < dest.x)
-          {
+        if (rect2.x < dest.x) {
             dest.width += dest.x - rect2.x;
             dest.x = rect2.x;
-          }
-        if (rect2.y < dest.y)
-          {
+        }
+        if (rect2.y < dest.y) {
             dest.height += dest.y - rect2.y;
             dest.y = rect2.y;
-          }
+        }
         if (rect2.x + rect2.width > dest.x + dest.width)
-          dest.width = rect2.x + rect2.width - dest.x;
+            dest.width = rect2.x + rect2.width - dest.x;
         if (rect2.y + rect2.height > dest.y + dest.height)
-          dest.height = rect2.y + rect2.height - dest.y;
+            dest.height = rect2.y + rect2.height - dest.y;
 
         return dest;
     }
@@ -54,10 +50,10 @@ class Rect {
     }
 
     overlap(rect2) {
-        return !((this.x + this.width    <= rect2.x) ||
-                 (rect2.x + rect2.width  <= this.x) ||
-                 (this.y + this.height   <= rect2.y) ||
-                 (rect2.y + rect2.height <= this.y));
+        return !(this.x + this.width    <= rect2.x ||
+                 rect2.x + rect2.width  <= this.x ||
+                 this.y + this.height   <= rect2.y ||
+                 rect2.y + rect2.height <= this.y);
     }
 
     center() {
@@ -68,19 +64,21 @@ class Rect {
         this.x += dx;
         this.y += dy;
     }
-};
+}
 
 class NaturalLayoutStrategy extends Workspace.LayoutStrategy {
-    constructor(settings) {
-        super();
+    constructor(params, settings) {
+        super(params);
         this._settings = settings;
     }
 
-    computeLayout(windows, layout) {
-        layout.windows = windows;
+    computeLayout(windows, _params) {
+        return {
+            windows,
+        };
     }
 
-    /**
+    /*
      * Returns clones with matching target coordinates and scales to arrange windows in a natural way that no overlap exists and relative window size is preserved.
      * This function is almost a 1:1 copy of the function
      * PresentWindowsEffect::calculateWindowTransformationsNatural() from KDE, see:
@@ -90,8 +88,8 @@ class NaturalLayoutStrategy extends Workspace.LayoutStrategy {
         // As we are using pseudo-random movement (See "slot") we need to make sure the list
         // is always sorted the same way no matter which window is currently active.
 
-        let area_rect = new Rect(area.x, area.y, area.width, area.height);
-        let bounds = area_rect.copy();
+        let areaRect = new Rect(area.x, area.y, area.width, area.height);
+        let bounds = areaRect.copy();
         let clones = layout.windows;
 
         let direction = 0;
@@ -106,36 +104,36 @@ class NaturalLayoutStrategy extends Workspace.LayoutStrategy {
             // This is used when the window is on the edge of the screen to try to use as much screen real estate as possible.
             directions[i] = direction;
             direction++;
-            if (direction == 4) {
+            if (direction === 4)
                 direction = 0;
-            }
         }
 
-        let loop_counter = 0;
+        let loopCounter = 0;
         let overlap;
         do {
             overlap = false;
             for (let i = 0; i < rects.length; i++) {
                 for (let j = 0; j < rects.length; j++) {
-                    if (i != j && rects[i].adjusted(-WINDOW_PLACEMENT_NATURAL_GAPS, -WINDOW_PLACEMENT_NATURAL_GAPS,
-                                                    WINDOW_PLACEMENT_NATURAL_GAPS, WINDOW_PLACEMENT_NATURAL_GAPS).overlap(
-                                                     rects[j].adjusted(-WINDOW_PLACEMENT_NATURAL_GAPS, -WINDOW_PLACEMENT_NATURAL_GAPS,
-                                                                       WINDOW_PLACEMENT_NATURAL_GAPS, WINDOW_PLACEMENT_NATURAL_GAPS))) {
-                        loop_counter++;
+                    let adjustments = [-1, -1, 1, 1]
+                        .map(v => (v *= WINDOW_PLACEMENT_NATURAL_GAPS));
+                    let iAdjusted = rects[i].adjusted(...adjustments);
+                    let jAdjusted = rects[j].adjusted(...adjustments);
+                    if (i !== j && iAdjusted.overlap(jAdjusted)) {
+                        loopCounter++;
                         overlap = true;
 
                         // TODO: something like a Point2D would be nicer here:
 
                         // Determine pushing direction
-                        let i_center = rects[i].center();
-                        let j_center = rects[j].center();
-                        let diff = [j_center[0] - i_center[0], j_center[1] - i_center[1]];
+                        let iCenter = rects[i].center();
+                        let jCenter = rects[j].center();
+                        let diff = [jCenter[0] - iCenter[0], jCenter[1] - iCenter[1]];
 
                         // Prevent dividing by zero and non-movement
-                        if (diff[0] == 0 && diff[1] == 0)
+                        if (diff[0] === 0 && diff[1] === 0)
                             diff[0] = 1;
                         // Try to keep screen/workspace aspect ratio
-                        if ( bounds.height / bounds.width > area_rect.height / area_rect.width )
+                        if (bounds.height / bounds.width > areaRect.height / areaRect.width)
                             diff[0] *= 2;
                         else
                             diff[1] *= 2;
@@ -164,33 +162,33 @@ class NaturalLayoutStrategy extends Workspace.LayoutStrategy {
                             let xSection = Math.round((rects[i].x - bounds.x) / (bounds.width / 3));
                             let ySection = Math.round((rects[i].y - bounds.y) / (bounds.height / 3));
 
-                            let i_center = rects[i].center();
+                            iCenter = rects[i].center();
                             diff[0] = 0;
                             diff[1] = 0;
-                            if (xSection != 1 || ySection != 1) { // Remove this if you want the center to pull as well
-                                if (xSection == 1)
-                                    xSection = (directions[i] / 2 ? 2 : 0);
-                                if (ySection == 1)
-                                    ySection = (directions[i] % 2 ? 2 : 0);
+                            if (xSection !== 1 || ySection !== 1) { // Remove this if you want the center to pull as well
+                                if (xSection === 1)
+                                    xSection = directions[i] / 2 ? 2 : 0;
+                                if (ySection === 1)
+                                    ySection = directions[i] % 2 ? 2 : 0;
                             }
-                            if (xSection == 0 && ySection == 0) {
-                                diff[0] = bounds.x - i_center[0];
-                                diff[1] = bounds.y - i_center[1];
+                            if (xSection === 0 && ySection === 0) {
+                                diff[0] = bounds.x - iCenter[0];
+                                diff[1] = bounds.y - iCenter[1];
                             }
-                            if (xSection == 2 && ySection == 0) {
-                                diff[0] = bounds.x + bounds.width - i_center[0];
-                                diff[1] = bounds.y - i_center[1];
+                            if (xSection === 2 && ySection === 0) {
+                                diff[0] = bounds.x + bounds.width - iCenter[0];
+                                diff[1] = bounds.y - iCenter[1];
                             }
-                            if (xSection == 2 && ySection == 2) {
-                                diff[0] = bounds.x + bounds.width - i_center[0];
-                                diff[1] = bounds.y + bounds.height - i_center[1];
+                            if (xSection === 2 && ySection === 2) {
+                                diff[0] = bounds.x + bounds.width - iCenter[0];
+                                diff[1] = bounds.y + bounds.height - iCenter[1];
                             }
-                            if (xSection == 0 && ySection == 2) {
-                                diff[0] = bounds.x - i_center[0];
-                                diff[1] = bounds.y + bounds.height - i_center[1];
+                            if (xSection === 0 && ySection === 2) {
+                                diff[0] = bounds.x - iCenter[0];
+                                diff[1] = bounds.y + bounds.height - iCenter[1];
                             }
-                            if (diff[0] != 0 || diff[1] != 0) {
-                                let length = Math.sqrt(diff[0]*diff[0] + diff[1]*diff[1]);
+                            if (diff[0] !== 0 || diff[1] !== 0) {
+                                length = Math.sqrt(diff[0] * diff[0] + diff[1] * diff[1]);
                                 diff[0] *= WINDOW_PLACEMENT_NATURAL_ACCURACY / length / 2;   // /2 to make it less influencing than the normal center-move above
                                 diff[1] *= WINDOW_PLACEMENT_NATURAL_ACCURACY / length / 2;
                                 rects[i].translate(diff[0], diff[1]);
@@ -203,78 +201,94 @@ class NaturalLayoutStrategy extends Workspace.LayoutStrategy {
                     }
                 }
             }
-        } while (overlap && loop_counter < WINDOW_PLACEMENT_NATURAL_MAX_TRANSLATIONS);
+        } while (overlap && loopCounter < WINDOW_PLACEMENT_NATURAL_MAX_TRANSLATIONS);
 
         // Work out scaling by getting the most top-left and most bottom-right window coords.
         let scale;
-        scale = Math.min(area_rect.width / bounds.width,
-                         area_rect.height / bounds.height,
-                         1.0);
+        scale = Math.min(
+            areaRect.width / bounds.width,
+            areaRect.height / bounds.height,
+            1.0);
 
         // Make bounding rect fill the screen size for later steps
-        bounds.x = bounds.x - (area_rect.width - bounds.width * scale) / 2;
-        bounds.y = bounds.y - (area_rect.height - bounds.height * scale) / 2;
-        bounds.width = area_rect.width / scale;
-        bounds.height = area_rect.height / scale;
+        bounds.x -= (areaRect.width - bounds.width * scale) / 2;
+        bounds.y -= (areaRect.height - bounds.height * scale) / 2;
+        bounds.width = areaRect.width / scale;
+        bounds.height = areaRect.height / scale;
 
         // Move all windows back onto the screen and set their scale
-        for (let i = 0; i < rects.length; i++) {
+        for (let i = 0; i < rects.length; i++)
             rects[i].translate(-bounds.x, -bounds.y);
-        }
 
-        // TODO: Implement the KDE part "Try to fill the gaps by enlarging windows if they have the space" here. (If this is wanted)
 
         // rescale to workspace
-        let scales = [];
-
-        let buttonOuterHeight, captionHeight;
-        let buttonOuterWidth = 0;
-
         let slots = [];
         for (let i = 0; i < rects.length; i++) {
-            rects[i].x = rects[i].x * scale + area_rect.x;
-            rects[i].y = rects[i].y * scale + area_rect.y;
+            rects[i].x = rects[i].x * scale + areaRect.x;
+            rects[i].y = rects[i].y * scale + areaRect.y;
+            rects[i].width *= scale;
+            rects[i].height *= scale;
 
-            slots.push([rects[i].x, rects[i].y, scale, clones[i]]);
+            slots.push([rects[i].x, rects[i].y, rects[i].width, rects[i].height, clones[i]]);
         }
 
         return slots;
     }
-};
+}
 
 let winInjections, workspaceInjections;
 
+/** */
 function resetState() {
     winInjections = { };
     workspaceInjections = { };
 }
 
+/** */
 function enable() {
     resetState();
 
-    let settings = Convenience.getSettings();
+    let settings = ExtensionUtils.getSettings();
 
-    workspaceInjections['_getBestLayout'] = Workspace.Workspace.prototype._getBestLayout;
-    Workspace.Workspace.prototype._getBestLayout = function(windows) {
-        let strategy = new NaturalLayoutStrategy(settings);
-        let layout = { strategy };
-        strategy.computeLayout(windows, layout);
+    workspaceInjections['_createBestLayout'] = Workspace.WorkspaceLayout.prototype._createBestLayout;
+    Workspace.WorkspaceLayout.prototype._createBestLayout = function (_area) {
+        this._layoutStrategy = new NaturalLayoutStrategy({
+            monitor: Main.layoutManager.monitors[this._monitorIndex],
+        }, settings);
+        return this._layoutStrategy.computeLayout(this._sortedWindows);
+    };
 
-        return layout;
-    }
+    // position window titles on top of windows in overlay
+    winInjections['_init'] = WindowPreview.prototype._init;
+    WindowPreview.prototype._init = function (...args) {
+        winInjections['_init'].call(this, ...args);
 
-    /// position window titles on top of windows in overlay ////
-    winInjections['relayout'] = Workspace.WindowOverlay.prototype.relayout;
-    Workspace.WindowOverlay.prototype.relayout = function(animate) {
-        if (settings.get_boolean('window-captions-on-top')) {
-            let [, , , cloneHeight] = this._windowClone.slot;
-            this.title.translation_y = -cloneHeight;
-        }
+        if (!settings.get_boolean('window-captions-on-top'))
+            return;
 
-        winInjections['relayout'].call(this, animate);
+        const alignConstraint = this._title.get_constraints().find(
+            c => c.align_axis && c.align_axis === Clutter.AlignAxis.Y_AXIS);
+        alignConstraint.factor = 0;
+
+        const bindConstraint = this._title.get_constraints().find(
+            c => c.coordinate && c.coordinate === Clutter.BindCoordinate.Y);
+        bindConstraint.offset = 0;
+    };
+    winInjections['_adjustOverlayOffsets'] =
+        WindowPreview.prototype._adjustOverlayOffsets;
+    WindowPreview.prototype._adjustOverlayOffsets = function (...args) {
+        winInjections['_adjustOverlayOffsets'].call(this, ...args);
+
+        if (settings.get_boolean('window-captions-on-top'))
+            this._title.translation_y = -this._title.translation_y;
     };
 }
 
+/**
+ * @param {Object} object - object that was modified
+ * @param {Object} injection - the map of previous injections
+ * @param {string} name - the @injection key that should be removed
+ */
 function removeInjection(object, injection, name) {
     if (injection[name] === undefined)
         delete object[name];
@@ -282,18 +296,15 @@ function removeInjection(object, injection, name) {
         object[name] = injection[name];
 }
 
+/** */
 function disable() {
     var i;
 
     for (i in workspaceInjections)
-        removeInjection(Workspace.Workspace.prototype, workspaceInjections, i);
+        removeInjection(Workspace.WorkspaceLayout.prototype, workspaceInjections, i);
     for (i in winInjections)
-        removeInjection(Workspace.WindowOverlay.prototype, winInjections, i);
+        removeInjection(WindowPreview.prototype, winInjections, i);
 
     global.stage.queue_relayout();
     resetState();
 }
-
-function init() {
-    /* do nothing */
-}
