@@ -11,8 +11,7 @@ const Me = ExtensionUtils.getCurrentExtension();
 const { WindowPicker, WindowPickerToggle } = Me.imports.windowPicker;
 const { WorkspaceIndicator } = Me.imports.workspaceIndicator;
 
-const Gettext = imports.gettext.domain(Me.metadata['gettext-domain']);
-const _ = Gettext.gettext;
+const _ = ExtensionUtils.gettext;
 
 const ICON_TEXTURE_SIZE = 24;
 const DND_ACTIVATE_TIMEOUT = 500;
@@ -23,34 +22,10 @@ const GroupingMode = {
     ALWAYS: 2,
 };
 
-
-function _minimizeOrActivateWindow(window) {
-    let focusWindow = global.display.focus_window;
-    if (focusWindow === window ||
-        focusWindow && focusWindow.get_transient_for() === window)
-        window.minimize();
-    else
-        window.activate(global.get_current_time());
-}
-
-function _openMenu(menu) {
-    menu.open();
-
-    let event = Clutter.get_current_event();
-    if (event && event.type() === Clutter.EventType.KEY_RELEASE)
-        menu.actor.navigate_focus(null, Gtk.DirectionType.TAB_FORWARD, false);
-}
-
-function _onMenuStateChanged(menu, isOpen) {
-    if (isOpen)
-        return;
-
-    let [x, y] = global.get_pointer();
-    let actor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE, x, y);
-    if (Me.stateObj.someWindowListContains(actor))
-        actor.sync_hover();
-}
-
+/**
+ * @param {Shell.App} app - an app
+ * @returns {number} - the smallest stable sequence of the app's windows
+ */
 function _getAppStableSequence(app) {
     let windows = app.get_windows().filter(w => !w.skip_taskbar);
     return windows.reduce((prev, cur) => {
@@ -58,7 +33,6 @@ function _getAppStableSequence(app) {
     }, Infinity);
 }
 
-
 class WindowContextMenu extends PopupMenu.PopupMenu {
     constructor(source, metaWindow) {
         super(source, 0.5, St.Side.BOTTOM);
@@ -282,10 +256,37 @@ const BaseButton = GObject.registerClass({
         return true;
     }
 
+    _openMenu(menu) {
+        menu.open();
+
+        let event = Clutter.get_current_event();
+        if (event && event.type() === Clutter.EventType.KEY_RELEASE)
+            menu.actor.navigate_focus(null, Gtk.DirectionType.TAB_FORWARD, false);
+    }
+
+    _minimizeOrActivateWindow(window) {
+        let focusWindow = global.display.focus_window;
+        if (focusWindow === window ||
+            focusWindow && focusWindow.get_transient_for() === window)
+            window.minimize();
+        else
+            window.activate(global.get_current_time());
+    }
+
+    _onMenuStateChanged(menu, isOpen) {
+        if (isOpen)
+            return;
+
+        let [x, y] = global.get_pointer();
+        let actor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE, x, y);
+        if (Me.stateObj.someWindowListContains(actor))
+            actor.sync_hover();
+    }
+
     _onPopupMenu(_actor) {
         if (!this._canOpenPopupMenu() || this._contextMenu.isOpen)
             return;
-        _openMenu(this._contextMenu);
+        this._openMenu(this._contextMenu);
     }
 
     _isFocused() {
@@ -362,7 +363,8 @@ class WindowButton extends BaseButton {
         this.label_actor = this._windowTitle.label_actor;
 
         this._contextMenu = new WindowContextMenu(this, this.metaWindow);
-        this._contextMenu.connect('open-state-changed', _onMenuStateChanged);
+        this._contextMenu.connect('open-state-changed',
+            this._onMenuStateChanged.bind(this));
         this._contextMenu.actor.hide();
         this._contextMenuManager.addMenu(this._contextMenu);
         Main.uiGroup.add_actor(this._contextMenu.actor);
@@ -382,9 +384,9 @@ class WindowButton extends BaseButton {
         }
 
         if (button === 1)
-            _minimizeOrActivateWindow(this.metaWindow);
+            this._minimizeOrActivateWindow(this.metaWindow);
         else
-            _openMenu(this._contextMenu);
+            this._openMenu(this._contextMenu);
     }
 
     _isFocused() {
@@ -518,14 +520,16 @@ class AppButton extends BaseButton {
 
         this._menuManager = new PopupMenu.PopupMenuManager(this);
         this._menu = new PopupMenu.PopupMenu(this, 0.5, St.Side.BOTTOM);
-        this._menu.connect('open-state-changed', _onMenuStateChanged);
+        this._menu.connect('open-state-changed',
+            this._onMenuStateChanged.bind(this));
         this._menu.actor.hide();
         this._menu.connect('activate', this._onMenuActivate.bind(this));
         this._menuManager.addMenu(this._menu);
         Main.uiGroup.add_actor(this._menu.actor);
 
         this._appContextMenu = new AppContextMenu(this);
-        this._appContextMenu.connect('open-state-changed', _onMenuStateChanged);
+        this._appContextMenu.connect('open-state-changed',
+            this._onMenuStateChanged.bind(this));
         this._appContextMenu.actor.hide();
         Main.uiGroup.add_actor(this._appContextMenu.actor);
 
@@ -592,7 +596,7 @@ class AppButton extends BaseButton {
                 this._singleWindowTitle.child = this._windowTitle;
                 this._windowContextMenu = new WindowContextMenu(this, this.metaWindow);
                 this._windowContextMenu.connect(
-                    'open-state-changed', _onMenuStateChanged);
+                    'open-state-changed', this._onMenuStateChanged.bind(this));
                 Main.uiGroup.add_actor(this._windowContextMenu.actor);
                 this._windowContextMenu.actor.hide();
                 this._contextMenuManager.addMenu(this._windowContextMenu);
@@ -631,7 +635,7 @@ class AppButton extends BaseButton {
             if (windows.length === 1) {
                 if (contextMenuWasOpen)
                     return;
-                _minimizeOrActivateWindow(windows[0]);
+                this._minimizeOrActivateWindow(windows[0]);
             } else {
                 this._menu.removeAll();
 
@@ -642,12 +646,12 @@ class AppButton extends BaseButton {
                     item._window = windows[i];
                     this._menu.addMenuItem(item);
                 }
-                _openMenu(this._menu);
+                this._openMenu(this._menu);
             }
         } else {
             if (contextMenuWasOpen)
                 return;
-            _openMenu(this._contextMenu);
+            this._openMenu(this._contextMenu);
         }
     }
 
@@ -761,10 +765,9 @@ class WindowList extends St.Widget {
 
         let workspaceManager = global.workspace_manager;
 
-        this._workspaceSignals = new Map();
         this._nWorkspacesChangedId = workspaceManager.connect(
-            'notify::n-workspaces', this._onWorkspacesChanged.bind(this));
-        this._onWorkspacesChanged();
+            'notify::n-workspaces', this._updateWorkspaceIndicatorVisibility.bind(this));
+        this._updateWorkspaceIndicatorVisibility();
 
         this._switchWorkspaceId = global.window_manager.connect(
             'switch-workspace', this._checkGrouping.bind(this));
@@ -774,16 +777,23 @@ class WindowList extends St.Widget {
             this._updateKeyboardAnchor();
         });
 
-        this._overviewHidingId = Main.overview.connect('hiding', () => {
+        this._overviewHidingId = Main.overview.connect('hidden', () => {
             this.visible = !Main.layoutManager.primaryMonitor.inFullscreen;
             this._updateKeyboardAnchor();
         });
 
         this._fullscreenChangedId =
             global.display.connect('in-fullscreen-changed', () => {
+                // Work-around for initial change from unknown to !fullscreen
+                if (Main.overview.visible)
+                    this.hide();
                 this._updateKeyboardAnchor();
             });
 
+        this._windowSignals = new Map();
+        this._windowCreatedId = global.display.connect(
+            'window-created', (dsp, win) => this._addWindow(win));
+
         this._dragBeginId = Main.xdndHandler.connect('drag-begin',
             this._monitorDrag.bind(this));
         this._dragEndId = Main.xdndHandler.connect('drag-end',
@@ -906,7 +916,7 @@ class WindowList extends St.Widget {
                        w2.metaWindow.get_stable_sequence();
             });
             for (let i = 0; i < windows.length; i++)
-                this._onWindowAdded(null, windows[i].metaWindow);
+                this._addWindow(windows[i].metaWindow);
         } else {
             let apps = this._appSystem.get_running().sort((a1, a2) => {
                 return _getAppStableSequence(a1) -
@@ -918,11 +928,8 @@ class WindowList extends St.Widget {
     }
 
     _updateKeyboardAnchor() {
-        if (!Main.keyboard.keyboardActor)
-            return;
-
-        let translationY = Main.overview.visible ? 0 : this.height;
-        Main.keyboard.keyboardActor.translation_y = -translationY;
+        const translationY = Main.overview.visible ? 0 : this.height;
+        Main.layoutManager.keyboardBox.translation_y = -translationY;
     }
 
     _onAppStateChanged(appSys, app) {
@@ -949,7 +956,7 @@ class WindowList extends St.Widget {
             child.destroy();
     }
 
-    _onWindowAdded(ws, win) {
+    _addWindow(win) {
         if (!this._grouped)
             this._checkGrouping();
 
@@ -960,21 +967,26 @@ class WindowList extends St.Widget {
         if (children.find(c => c.metaWindow === win))
             return;
 
+        this._windowSignals.set(
+            win, win.connect('unmanaged', () => this._removeWindow(win)));
+
         let button = new WindowButton(win, this._perMonitor, this._monitor.index);
         this._settings.bind('display-all-workspaces',
             button, 'ignore-workspace', Gio.SettingsBindFlags.GET);
         this._windowList.add_child(button);
     }
 
-    _onWindowRemoved(ws, win) {
+    _removeWindow(win) {
         if (this._grouped)
             this._checkGrouping();
 
         if (this._grouped)
             return;
 
-        if (win.get_compositor_private())
-            return; // not actually removed, just moved to another workspace
+        const id = this._windowSignals.get(win);
+        if (id)
+            win.disconnect(id);
+        this._windowSignals.delete(id);
 
         let children = this._windowList.get_children();
         let child = children.find(c => c.metaWindow === win);
@@ -982,39 +994,6 @@ class WindowList extends St.Widget {
             child.destroy();
     }
 
-    _onWorkspacesChanged() {
-        let workspaceManager = global.workspace_manager;
-        let numWorkspaces = workspaceManager.n_workspaces;
-
-        for (let i = 0; i < numWorkspaces; i++) {
-            let workspace = workspaceManager.get_workspace_by_index(i);
-            if (this._workspaceSignals.has(workspace))
-                continue;
-
-            let signals = { windowAddedId: 0, windowRemovedId: 0 };
-            signals._windowAddedId = workspace.connect_after(
-                'window-added', this._onWindowAdded.bind(this));
-            signals._windowRemovedId = workspace.connect(
-                'window-removed', this._onWindowRemoved.bind(this));
-            this._workspaceSignals.set(workspace, signals);
-        }
-
-        this._updateWorkspaceIndicatorVisibility();
-    }
-
-    _disconnectWorkspaceSignals() {
-        let workspaceManager = global.workspace_manager;
-        let numWorkspaces = workspaceManager.n_workspaces;
-
-        for (let i = 0; i < numWorkspaces; i++) {
-            let workspace = workspaceManager.get_workspace_by_index(i);
-            let signals = this._workspaceSignals.get(workspace);
-            this._workspaceSignals.delete(workspace);
-            workspace.disconnect(signals._windowAddedId);
-            workspace.disconnect(signals._windowRemovedId);
-        }
-    }
-
     _monitorDrag() {
         DND.addDragMonitor(this._dragMonitor);
     }
@@ -1078,18 +1057,20 @@ class WindowList extends St.Widget {
         Main.keyboard._bottomDragAction.disconnect(this._keyboardVisiblechangedId);
         this._keyboardVisiblechangedId = 0;
 
-        this._disconnectWorkspaceSignals();
         global.workspace_manager.disconnect(this._nWorkspacesChangedId);
         this._nWorkspacesChangedId = 0;
 
         global.window_manager.disconnect(this._switchWorkspaceId);
         this._switchWorkspaceId = 0;
 
+        this._windowSignals.forEach((id, win) => win.disconnect(id));
+        this._windowSignals.clear();
 
         Main.overview.disconnect(this._overviewShowingId);
         Main.overview.disconnect(this._overviewHidingId);
 
         global.display.disconnect(this._fullscreenChangedId);
+        global.display.disconnect(this._windowCreatedId);
 
         this._stopMonitoringDrag();
         Main.xdndHandler.disconnect(this._dragBeginId);
@@ -1170,6 +1151,9 @@ class Extension {
     }
 }
 
+/**
+ * @returns {Extension} - the extension's state object
+ */
 function init() {
     return new Extension();
 }
