@@ -1,121 +1,126 @@
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 /* exported init */
-const { Clutter, St } = imports.gi;
+const { Clutter, Graphene, GObject, St } = imports.gi;
 
 const Main = imports.ui.main;
+const OverviewControls = imports.ui.overviewControls;
 const Workspace = imports.ui.workspace;
 const WorkspacesView = imports.ui.workspacesView;
 
-var MyWindowOverlay = class extends Workspace.WindowOverlay {
-    constructor(windowClone, parentActor) {
-        super(windowClone, parentActor);
+const WINDOW_SLOT = 4;
 
-        this._id = null;
-        this._text = new St.Label({
-            style_class: 'extension-windowsNavigator-window-tooltip',
-            visible: false,
-        });
-        parentActor.add_actor(this._text);
-    }
+var MyWorkspace = GObject.registerClass(
+class MyWorkspace extends Workspace.Workspace {
+    _init(...args) {
+        super._init(...args);
 
-    showTooltip() {
-        this._text.raise_top();
-        this._text.show();
-        this._text.text = (this._windowClone.slotId + 1).toString();
-    }
-
-    hideTooltip() {
-        if (this._text && this._text.visible)
-            this._text.hide();
-    }
-
-    relayout(animate) {
-        super.relayout(animate);
-
-        let [cloneX, cloneY, cloneWidth_, cloneHeight_] = this._windowClone.slot;
-
-        let textX = cloneX - 2;
-        let textY = cloneY - 2;
-        this._text.set_position(Math.floor(textX) + 5, Math.floor(textY) + 5);
-        this._text.raise_top();
-    }
-};
-
-var MyWorkspace = class extends Workspace.Workspace {
-    constructor(metaWorkspace, monitorIndex) {
-        super(metaWorkspace, monitorIndex);
-
-        if (metaWorkspace && metaWorkspace.index() < 9) {
+        if (this.metaWorkspace && this.metaWorkspace.index() < 9) {
             this._tip = new St.Label({
                 style_class: 'extension-windowsNavigator-window-tooltip',
                 visible: false,
             });
-            this.actor.add_actor(this._tip);
+            this.add_actor(this._tip);
 
-            this.actor.connect('notify::scale-x', () => {
-                this._tip.set_scale(1 / this.actor.scale_x, 1 / this.actor.scale_x);
+            this.connect('notify::scale-x', () => {
+                this._tip.set_scale(1 / this.scale_x, 1 / this.scale_x);
             });
         } else {
             this._tip = null;
         }
     }
 
+    vfunc_allocate(box) {
+        super.vfunc_allocate(box);
+
+        if (this._tip)
+            this._tip.allocate_preferred_size(0, 0);
+    }
+
     showTooltip() {
-        if (!this._tip || !this._actualGeometry)
+        if (!this._tip)
             return;
         this._tip.text = (this.metaWorkspace.index() + 1).toString();
-
-        // Hand code this instead of using _getSpacingAndPadding
-        // because that fails on empty workspaces
-        let node = this.actor.get_theme_node();
-        let padding = {
-            left: node.get_padding(St.Side.LEFT),
-            top: node.get_padding(St.Side.TOP),
-            bottom: node.get_padding(St.Side.BOTTOM),
-            right: node.get_padding(St.Side.RIGHT),
-        };
-
-        let area = Workspace.padArea(this._actualGeometry, padding);
-        this._tip.x = area.x;
-        this._tip.y = area.y;
         this._tip.show();
-        this._tip.raise_top();
+        this.set_child_below_sibling(this._tip, null);
     }
 
     hideTooltip() {
-        if (!this._tip)
-            return;
-        if (!this._tip.get_parent())
-            return;
-        this._tip.hide();
+        if (this._tip)
+            this._tip.hide();
     }
 
     getWindowWithTooltip(id) {
-        for (let i = 0; i < this._windows.length; i++) {
-            if (this._windows[i].slotId + 1 === id)
-                return this._windows[i].metaWindow;
-        }
-        return null;
+        const { layoutManager } = this._container;
+        const slot = layoutManager._windowSlots[id - 1];
+        return slot ? slot[WINDOW_SLOT].metaWindow : null;
     }
 
     showWindowsTooltips() {
-        for (let i in this._windowOverlays) {
-            if (this._windowOverlays[i])
-                this._windowOverlays[i].showTooltip();
+        const { layoutManager } = this._container;
+        for (let i = 0; i < layoutManager._windowSlots.length; i++) {
+            if (layoutManager._windowSlots[i])
+                layoutManager._windowSlots[i][WINDOW_SLOT].showTooltip(`${i + 1}`);
         }
     }
 
     hideWindowsTooltips() {
-        for (let i in this._windowOverlays) {
-            if (this._windowOverlays[i])
-                this._windowOverlays[i].hideTooltip();
+        const { layoutManager } = this._container;
+        for (let i in layoutManager._windowSlots) {
+            if (layoutManager._windowSlots[i])
+                layoutManager._windowSlots[i][WINDOW_SLOT].hideTooltip();
         }
     }
-};
 
-var MyWorkspacesView = class extends WorkspacesView.WorkspacesView {
-    constructor(width, height, x, y, workspaces) {
-        super(width, height, x, y, workspaces);
+    // overriding _addWindowClone to apply the tooltip patch on the cloned
+    // windowPreview
+    _addWindowClone(metaWindow) {
+        const clone = super._addWindowClone(metaWindow);
+
+        // appling the tooltip patch
+        (function patchPreview() {
+            this._text = new St.Label({
+                style_class: 'extension-windowsNavigator-window-tooltip',
+                visible: false,
+            });
+
+            this._text.add_constraint(new Clutter.BindConstraint({
+                source: this.windowContainer,
+                coordinate: Clutter.BindCoordinate.POSITION,
+            }));
+            this._text.add_constraint(new Clutter.AlignConstraint({
+                source: this.windowContainer,
+                align_axis: Clutter.AlignAxis.X_AXIS,
+                pivot_point: new Graphene.Point({ x: 0.5, y: -1 }),
+                factor: this._closeButtonSide === St.Side.LEFT ? 1 : 0,
+            }));
+            this._text.add_constraint(new Clutter.AlignConstraint({
+                source: this.windowContainer,
+                align_axis: Clutter.AlignAxis.Y_AXIS,
+                pivot_point: new Graphene.Point({ x: -1, y: 0.5 }),
+                factor: 0,
+            }));
+
+            this.add_child(this._text);
+        }).call(clone);
+
+        clone.showTooltip = function (text) {
+            this._text.set({ text });
+            this._text.show();
+        };
+
+        clone.hideTooltip = function () {
+            if (this._text && this._text.visible)
+                this._text.hide();
+        };
+
+        return clone;
+    }
+});
+
+var MyWorkspacesView = GObject.registerClass(
+class MyWorkspacesView extends WorkspacesView.WorkspacesView {
+    _init(...args) {
+        super._init(...args);
 
         this._pickWorkspace = false;
         this._pickWindow = false;
@@ -159,8 +164,8 @@ var MyWorkspacesView = class extends WorkspacesView.WorkspacesView {
     }
 
     _onKeyPress(s, o) {
-        let { viewSelector } = Main.overview;
-        if (viewSelector._activePage !== viewSelector._workspacesPage)
+        const { ControlsState } = OverviewControls;
+        if (this._overviewAdjustment.value !== ControlsState.WINDOW_PICKER)
             return false;
 
         let workspaceManager = global.workspace_manager;
@@ -238,28 +243,28 @@ var MyWorkspacesView = class extends WorkspacesView.WorkspacesView {
         }
         return false;
     }
-};
+});
 
 class Extension {
     constructor() {
-        this._origWindowOverlay = Workspace.WindowOverlay;
         this._origWorkspace = Workspace.Workspace;
         this._origWorkspacesView = WorkspacesView.WorkspacesView;
     }
 
     enable() {
-        Workspace.WindowOverlay = MyWindowOverlay;
         Workspace.Workspace = MyWorkspace;
         WorkspacesView.WorkspacesView = MyWorkspacesView;
     }
 
     disable() {
-        Workspace.WindowOverlay = this._origWindowOverlay;
         Workspace.Workspace = this._origWorkspace;
         WorkspacesView.WorkspacesView = this._origWorkspacesView;
     }
 }
 
+/**
+ * @returns {Extension} - the extension's state object
+ */
 function init() {
     return new Extension();
 }
